/// \file mil1553ud_ioctl.h
/// \brief инфраструктура ioctl-команд
#ifndef MIL1553UD_CMD_H
#define MIL1553UD_CMD_H

/// \remark ioctl-команды драйвера

#define IOC_MAGIC 'b'
// read/write regs
/// \brief команда "запись в регистр"
/// позволяет записть значение в заданный регистр управления
#define IOCTL_WRITE_REG                             _IOW(IOC_MAGIC, 0,  SADDR_DATA)             // запись в регистр
/// \brief команда "запись в регистр заданных бит"
/// позволяет изменить отдельные биты в заданном регистре управления
#define IOCTL_WRITE_REG_BIT_MASK                    _IOW(IOC_MAGIC, 1,  SADDR_DATA_BIT_MASK)    // запись в регистр заданных бит
/// \brief команда "чтение из регистра"
/// позволяет прочитать значение из заданного регистра управления
#define IOCTL_READ_REG                              _IOWR(IOC_MAGIC,2,  SADDR_DATA)             // чтение из регитсра
// info command
/// \brief команда "чтение подробной информации о плате и драйвере"
/// позволяет считать подробную информацию о pci-плате
#define IOCTL_VERSION                               _IOR(IOC_MAGIC, 31, VERSION)                // получить информацию  о плате
/// \brief команда "чтение версии драйвера"
/// позволяет считать версию и дату драйвера
#define IOCTL_VERSION_DRIVER                        _IOR(IOC_MAGIC, 40, unsigned int)           // получить информацию о драйвере
/// \brief команда "чтение информации о pci-локации платы"
/// позволяет считать информацию pci
#define IOCTL_PCI_LOCATION                          _IOR(IOC_MAGIC, 41, PCI_LOCATION)           // получить информацию о pci-локации платы
// dma
/// \brief команда "сброс указателя dma канала"
/// позволяет обнулить DMA_INDEX и программные указатели чтения/записи
#define IOCTL_CLEAR_DMA                             _IO(IOC_MAGIC,  5)                          // сброс указателя dma канала
/// \brief команда "включить dma устройства"
/// разрешает работу DMA
#define IOCTL_ENABLE_DMA                            _IO(IOC_MAGIC,  8)                          // включить dma платы
/// \brief команда "выключить dma устройства"
/// запрещает работу DMA
#define IOCTL_DISABLE_DMA                           _IO(IOC_MAGIC,  9)                          // выключить dma платы
/// \brief команда "получить количество 128 байтных блоков из ДМА"
/// позволяет получить количество готовых для чтения блоков из буфера ДМА
#define IOCTL_GET_NBLOCK_RAW_DMA                    _IOWR(IOC_MAGIC,7,  unsigned int)           // получить количество блоков из ДМА в сыром виде
/// \brief команда "считать заданное количество 128 байтных из ДМА"
/// позволяет получить данные (блоки) из буфера ДМА
#define IOCTL_READ_BLOCKS_RAW_DMA                   _IOWR(IOC_MAGIC,6,  DMA_READ_BLOCK)         // считать заданное кол-во блоков из ДМА
#define IOCTL_READ_BLOCKS_RAW_DMA_FIX_COUNT         _IOWR(IOC_MAGIC,60, MIL1553_DMA_FIX_BLOCK)      // считать фиксированное кол-во блоков из ДМА (не более 256 за раз)
// irq
/// \brief команда "управление прерываниями INTERRUPT_MASK"
#define IOCTL_MANAGE_INTERRUPT                      _IOW(IOC_MAGIC, 16, INTERRUPT_MAN)          // управление прерываниями INTERRUPT_MASK
/// \brief команда "управление прерываниями по передаче из подадреса"
#define IOCTL_MAN_IRQ_SUBADDRESS_TR                 _IOW(IOC_MAGIC, 38, INTERRUPT_SUBADDRESS)   // управление прерываниями по передаче из подадреса
/// \brief команда "управление прерываниями по приёму в подадрес"
#define IOCTL_MAN_IRQ_SUBADDRESS_RCV                _IOW(IOC_MAGIC, 39, INTERRUPT_SUBADDRESS)   // управление прерываниями по приёму в подадрес
// mode
/// \brief команда "выбор режима работы канала"
#define IOCTL_SWITCH_MODE                           _IOW(IOC_MAGIC, 4,  unsigned int)           // выбор режима работы канала
/// \brief команда "выбор шины канала"
#define IOCTL_SWITCH_BUS                            _IOW(IOC_MAGIC, 33, unsigned int)           // выбор шины канала
/// \brief команда "разрешить работу канала"
#define IOCTL_DEV_CHANNEL_WORK                      _IOW(IOC_MAGIC, 34, unsigned int)           // вкл/выкл канал
// term dev
/// \brief команда "установить режим ОУ"
#define IOCTL_SET_RT_TR_MODE                        _IOW(IOC_MAGIC, 35, RT_TR_MODE)             // установить режим ОУ
/// \brief команда "установить готовность буфера ОУ"
#define IOCTL_SET_RT_TR_BUF_READY                   _IOW(IOC_MAGIC, 37, RT_TR_MODE_BUF)         // установить готовность буфера ОУ
/// \brief команда "управление разрешением буферов"
#define IOCTL_RT_BUF_MAN_EN                         _IOW(IOC_MAGIC, 36, RT_BUF_MAN_EN)          // установка разрешение буферов ОУ
// channel
/// \brief команда "установить адрес ОУ"
#define IOCTL_SET_ADDRESS                           _IOW(IOC_MAGIC, 19, unsigned char)          // установить адрес ОУ
/// \brief команда "установить таймауты ОУ"
#define IOCTL_SET_TIMER_RTBM_CONF_REG_PCI           _IOW(IOC_MAGIC, 20, TIMER_TIMEOUT)          // set timer and timeout
/// \brief команда "установить таймауты КШ"
#define IOCTL_SET_TIMER_BC_CONF_REG_PCI             _IOW(IOC_MAGIC, 32, TIMER_TIMEOUT)          // set timer and timeout
// paddr and buf
/// \brief команда "записать подадрес на отправку"
#define IOCTL_WR_BLOCK_BUF_SUBADR                   _IOW(IOC_MAGIC, 27, RT_TR_BUF_SUBADDR)      // записать поддадрес
/// \brief команда "прочитать подадрес на отправку"
#define IOCTL_RD_BLOCK_BUF_SUBADR                   _IOWR(IOC_MAGIC,28, RT_TR_BUF_SUBADDR)      // прочитать поддадрес
// bc_ram
/// \brief команда "запись блока в BC RAM"
#define IOCTL_WRITE_BC_RAM                          _IOW(IOC_MAGIC, 29, BC_RAM_BLOCK)           // запись в BC ram
/// \brief команда "чтение блока из BC RAM"
#define IOCTL_READ_BC_RAM                           _IOWR(IOC_MAGIC,30, BC_RAM_BLOCK)           // чтение из BC ram
// interrupt ring buffer
/// \brief получить кол-во блоков с информацией о возникших прерываниях из кольцевого буфера
#define IOCTL_GET_COUNT_INTERRUPT_BLOCKS            _IOWR(IOC_MAGIC,50, unsigned int)           // получить кол-во блоков с информацией о возникших прерываниях
/// \brief считать заданное кол-во блоков из кольцевого буфера
#define IOCTL_READ_INTERRUPT_BLOCKS                 _IOWR(IOC_MAGIC,51, INTERRUPT_BLOCK_BUFFER) // считать заданное кол-во блоков из кольцевого буфера
/// \brief считать входной поддадрес
#define IOCTL_GET_RCV_SUBADR_DATA                   _IOWR(IOC_MAGIC,52, SUBA_DATA_BLOCK)        // считать входной поддадрес
/// \brief получить номер канала
#define IOCTL_GET_NUMBER_CH                         _IOWR(IOC_MAGIC,55,  unsigned int)          // получить номер канала
// signals
//#define IOCTL_SET_SIGNAL_IRQ                        _IOW(IOC_MAGIC, 53, SIGNAL_INFO)            // установить сигнал для прерывания
//#define IOCTL_CLR_SIGNAL_IRQ                        _IO(IOC_MAGIC,  54)


///
/// \remark КОНСТАНТЫ ДЛЯ IOCTL-команда
///

/// \remark константы для управления прерываниями
/// 0 - не меняется значение, 1 - включить прерывание, 2 - выключить прерывание
/// \brief не изменять бит прерывания
#define INTERRUPT_MAN_NO_CHANGE         0
/// \brief включить бит прерывания
#define INTERRUPT_MAN_ON                1
/// \brief выключить бит прерывания
#define INTERRUPT_MAN_OFF               2

/// \remark константы режимов работы устройства
/// \brief адресуемый монитор шины
#define MIL_MODE_MONITOR_ADRR           0x0
/// \brief контроллер шины
#define MIL_MODE_BUS_CONTR              0x1
/// \brief оконечное устройство
#define MIL_MODE_TERMINAL_DEV           0x2
/// \brief НЕадресуемый монитор шины
#define MIL_MODE_MONITOR                0x3

/// \remark константы выбора шины канала
/// \brief включить шину А
#define MIL_BUS_A_EN                    0x1
/// \brief включить шину В
#define MIL_BUS_B_EN                    0x2

/// \remark константы вкл/выкл канала
/// \brief включить канал в работу
#define MIL_DEV_CHANNEL_ON              0x1
/// \brief выключить канал
#define MIL_DEV_CHANNEL_OFF             0x0

/// \remark константы для режима ОУ
/// \brief режим - программный
#define MIL_RT_MODE_PROG                0x0
/// \brief режим - аппаратный
#define MIL_RT_MODE_HRDW                0x1

/// \remark команды управления буферами cmd_buf
/// \brief буфер данных RTF_BUF0 и RTF_BUF1 - выкл
#define MIL_RT_BUF_0_OFF_1_OFF          0x0
/// \brief буфер данных RTF_BUF0 - выкл и RTF_BUF1 - вкл
#define MIL_RT_BUF_0_OFF_1_ON           0x1
/// \brief буфер данных RTF_BUF0 - вкл и RTF_BUF1 - выкл
#define MIL_RT_BUF_0_ON_1_OFF           0x2
/// \brief буфер данных RTF_BUF0 - вкл и RTF_BUF1 - вкл
#define MIL_RT_BUF_0_ON_1_ON            0x3

/// \remark направление передачи direction
/// \brief направление - передача
#define MIL_RT_BUF_TRANSMIT             0x0
/// \brief направление - передача
#define MIL_RT_BUF_RECEIVE              0x1

/// \remark номер буфера num_buf
/// \brief отправной буфер данных RTF_BUF0
#define MIL_RT_BUF0                     0
/// \brief отправной буфер данных RTF_BUF1
#define MIL_RT_BUF1                     1

/// \remark константы для timer timeout RCVCK
#define MIL_T_RCVCK_17MKS               0x0
#define MIL_T_RCVCK_60MKS               0x1
#define MIL_T_RCVCK_85MKS               0x2
#define MIL_T_RCVCK_110MKS              0x3

/// \remark константы для timer timeout TRCK
#define MIL_T_TRCK_6MKS                 0x0
#define MIL_T_TRCK_8MKS                 0x1
#define MIL_T_TRCK_11MKS                0x2
#define MIL_T_TRCK_13MKS                0x3
#define MIL_T_TRCK_18MKS                0x4
#define MIL_T_TRCK_61MKS                0x5
#define MIL_T_TRCK_86MKS                0x6
#define MIL_T_TRCK_111MKS               0x7

/// \remark константы для timer timeout TCK
#define MIL_T_TCK_OFF                   0x0
#define MIL_T_TCK_1MKS                  0x1
#define MIL_T_TCK_2MKS                  0x2
#define MIL_T_TCK_4MKS                  0x3
#define MIL_T_TCK_8MKS                  0x4
#define MIL_T_TCK_16MKS                 0x5
#define MIL_T_TCK_32MKS                 0x6
#define MIL_T_TCK_64MKS                 0x7

/// \remark константы для команды записи/чтения BC RAM
#define MIL_BC_RAM_TYPE_INSTRUCTION     0x0
#define MIL_BC_RAM_TYPE_OPERATION       0x1
#define MIL_BC_RAM_TYPE_DATA            0x2

/// \remark номера каналов
/// \brief канал 1
#define CH_NUM_1            0
/// \brief канал 2
#define CH_NUM_2            1
/// \brief канал 3
#define CH_NUM_3            2
/// \brief канал 4
#define CH_NUM_4            3

/// \brief блок информации о прерывании
struct block_info {
    /// \brief значение таймера в момент прерывани
    unsigned int    free_timer_value;
    /// \brief маска прерываний (собственно по ней поймём какие в этот момент времени были прерывания)
    /// 0 - INT_HDAT, 1 - INT_QDAT, 2 - RT_INT_SADDR, 3 - RT_INT_MC_ERR,
    /// 4 - INT_BC, 5 - INT_FLASH, 6 - INT_DATA_CNT_EN,
    /// 7 - INT_TIMEOUT_ITVEN, 8 - INT_TIMEOUT_ABSEN
    unsigned int    interrupt_ch;
    /// \brief регистр HW_STAT_REG1
    unsigned int    hw_stat_reg1;
    /// \brief регистр HW_STAT_REG2
    unsigned int    hw_stat_reg2;
};

#define BLINF_INT_HDAT          0
#define BLINF_INT_QDAT          1
#define BLINF_RT_INT_SADDR      2
#define BLINF_RT_INT_MC_ERR     3
#define BLINF_INT_BC            4
#define BLINF_INT_FLASH         5
#define BLINT_INT_DATA_CNT_EN   6
#define BLINT_TIMEOUT_ITVEN     7
#define BLINT_TIMEOUT_ABSEN     8

#define SIZE_OF_PCI_LOC         13

///
/// \remark ТИПЫ ДАННЫХ АРГУМЕНТОВ IOCTL-команд
///

#define MIL1553_ONE_BLOCK_SIZE_128_BYTES        128

/// \brief контейнер блоков ДМА (на 256 блоков - максимум)
typedef struct {
    unsigned int currentBlocks;      ///< фактическое количество блоков ДМА (не более mallocedBlocks)
    unsigned int nextBlocks;         ///< кол-во блоков, готовых к чтению в целевом устройстве (кольцевом буфере ДМА девайса)
    unsigned int mallocedBlocks;     ///< выделенное место в data
    unsigned char* data;    ///< блоки ДМА в сыром виде (кратные 128 байтам)
} MIL1553_DMA_FIX_BLOCK;

/// \brief информация о локации на шине pci
typedef struct {
    ///
    /// \brief name инфо о локации
    /// например, 0000:00:02.0 -
    /// выделены домен (16 бит), шина (8 бит), устройство (5 бит) и функция (3 бита)
    ///
    char name[SIZE_OF_PCI_LOC];
} PCI_LOCATION;

/// \brief информация о драйвере и устройстве
typedef struct {
    /// \brief идентификатор устройства
    unsigned int device_id;
    /// \brief вендор устройства
    unsigned int vendor_id;
    /// \brief тип устройства (кол-во каналов)
    unsigned int type;
    /// \brief ревизия устройства
    char revision;
    /// \brief имя символьного устройства
    char dev_name[30];
    /// \brief значение минора
    int minor;
    /// \brief номер прерывания
    int irq;
    /// \brief размер ДМА буфера
    long size_dma;
    /// \brief виртуальный адрес ДМА буфера
    void* addr_dma_virt;
    /// \brief адрес bar-пространства
    unsigned int pci_bars;
    /// \brief виртуальный адрес bar-пространства
    void* addr_bar_virt;
} VERSION;

/// \brief запись/чтение регистров
typedef struct {
    /// \brief адрес регистра (смещение в соотв. со спецификацией)
    unsigned long daddr;
    /// \brief значение регистра
    unsigned int data;
} SADDR_DATA;

/// \brief запись в регистр заданных бит
typedef struct {
    /// \brief адрес регистра (смещение в соотв. со спецификацией)
    unsigned long daddr;
    /// \brief значение регистра
    unsigned int data;
    /// \brief битовая маска
    /// 1 - бит записывается из data, 0 - бит остаётся неизменным
    unsigned int mask;
} SADDR_DATA_BIT_MASK;

/// \brief считываемый блок дма, кратный 128 байтам
typedef struct {
    /// \brief количество 128 байтных блоков
    unsigned int countBlocks;
    /// \brief длина data в байтах
    unsigned int length;
    /// \brief данные блоков в сыром виде
    unsigned char* data;
} DMA_READ_BLOCK;

/// \remark типы данных для работы с прервываниями

/// \brief управление маскированием перываний поканально
typedef struct {
    /// \brief состояние прерывания от контроллера flash
    unsigned char int_flash;
    /// \brief состояние прерывания от КШ
    unsigned char int_bc;
    /// \brief состояние прерывания от ОУ при приёме данных
    unsigned char int_rt_ren;
    /// \brief состояние прерывания от ОУ при отправке данных
    unsigned char int_rt_ten;
    /// \brief состояние прерывания при заполнении 1/18
    unsigned char int_qdat;
    /// \brief состояние прерывания при заполнении 1/2
    unsigned char int_hdat;
    /// \brief состояние прерывания счётчика данных контроллера MIL
    unsigned char int_data_cnt_en;
    /// \brief состояние прерывания интервального таймера MIL
    unsigned char int_timeout_itven;
    /// \brief состояние прерывания абсолютного таймера MIL
    unsigned char int_timeout_absen;
} INTERRUPT_MAN;

/// \brief управление маскированием прерываний по приёму/передачи данных по подадресам
typedef struct {
    /// \brief состояние прерывания (вкл - 1 / выкл - 2)
    unsigned char rt_int;
    /// \brief маскирование прерываний по подадресам 1й бит - 1й подадрес и т.д.
    unsigned int subaddress_mask;
} INTERRUPT_SUBADDRESS;

/// \remark типы данных для работы с режимами

/// \brief управление режимом работы ОУ по подадресам
typedef struct {
    /// \brief подадрес
    unsigned char subaddress;
    /// \brief режим
    unsigned int mode;
} RT_TR_MODE;

/// \brief управление буферами 1 и 2
typedef struct {
    /// \brief подадрес 1 - 30
    unsigned char subaddress;
    /// \brief команда упр. буфером
    unsigned int cmd_buf;
} RT_TR_MODE_BUF;

/// \brief управление подадресами в режиме ОУ
typedef struct {
    /// \brief направление (приём/передача)
    unsigned int direction;
    /// \brief маскирование (выбор) по подадресам 1й бит - 1й подадрес и т.д.
    /// 1 - бит изменяем, 0 - бит не изменяем
    unsigned int subaddress_mask;
    /// \brief маскирование (выбор) действий по подадресам 1й бит - 1й подадрес и т.д.
    /// 1 - вкл, 0 - выкл
    unsigned int action_mask;
} RT_BUF_MAN_EN;

/// \brief буфер ОУ на передачу подадреса
typedef struct {
    /// \brief номер буфера
    unsigned char num_buf;
    /// \brief подадрес (1-30)
    unsigned char subaddress;
    /// \brief подадрес (32 слова данных)
    unsigned short data_words[32];
} RT_TR_BUF_SUBADDR;

/// \brief управление таймаутами таймеров
typedef struct {
    /// \brief значение RT_TCK
    unsigned int tck;
    /// \brief значение RT_TRCK
    unsigned int trck;
    /// \brief значение RT_RCVCK
    unsigned int rcvck;
} TIMER_TIMEOUT;

/// \brief блок данных для записи в BC_RAM (область микропрограммы для КШ)
typedef struct {
    /// \brief тип инструкций
    /// INSTR, OPERATION, DATA
    unsigned int type;
    /// \brief смещение от начала буфера в 32-х разрядных словах
    /// INSTR - max 4095, OPERATION - max 4095, DATA - max 8191
    unsigned int shift;
    /// \brief размер поля dwords в 32-х разрядных словах
    unsigned int length;
    /// \brief данные (массив 32-х разрядных слов)
    unsigned int* dwords;
} BC_RAM_BLOCK;

/// \brief блок данных для чтения накопленных прерываний
typedef struct {
    /// \brief кол-во запришиваемых блоков
    unsigned int        count_blocks;
    /// \brief собственно указатель на массив блоков
    struct block_info*  blocks;
} INTERRUPT_BLOCK_BUFFER;

/// \brief информация о подадресе
typedef struct {
    /// \brief подадрес (1-30)
    unsigned char sa;
    /// \brief слова данных
    unsigned short words[32];
} SUBA_DATA_BLOCK;

/// \brief информация для генерации сигнала
typedef struct {
    /// \brief pid целевого процесса
    /// кому посылать
    int target_pid;
    /// \brief используемый сигнал
    int sig_number;
}
SIGNAL_INFO;

/// \remark КОДЫ ОШИБОК IOCTL-запросов

/// \brief нет ошибок
#define GOOD                            0
/// \brief ошибка
#define ERROR                           -1000
/// \brief ошибка адреса регистра
#define ERR_BAD_ADDRESS                 -1001
/// \brief ошибка номера канала
/// внутрення ошибка в драйвере
#define ERR_BAD_CHANNEL                 -1002
/// \brief ошибка копирования данных в юзерспейс
#define ERR_COPY_TO_USER                -1003
/// \brief ошибка аргумента - mode
#define ERR_BAD_ARG_MODE                -1004
/// \brief ошибка аргумента - subaddress
#define ERR_BAD_ARG_SA                  -1005
/// \brief ошибка аргумента - direction
#define ERR_BAD_ARG_DIRECTION           -1006
/// \brief ошибка аргумента - tck
#define ERR_BAD_ARG_TCK                 -1007
/// \brief ошибка аргумента - trck
#define ERR_BAD_ARG_TRCK                -1008
/// \brief ошибка аргумента - rcvck
#define ERR_BAD_ARG_RCVCK               -1009
/// \brief ошибка аргумента - numbuf
#define ERR_BAD_ARG_NUMBUF              -1010
/// \brief ошибка аргумента - typr
#define ERR_BAD_ARG_TYPE                -1011
/// \brief ошибка выделения памяти
/// внутренняя ошибка драйвера
#define ERR_BAD_ALLOC_MEM               -1012
#define ERR_BAD_DMA_INDEX               -1013
#define ERR_BAD_DMA_NBLOCK              -1014


#endif // MIL1553UD_CMD_H
