/// \file mil1553ud_info.h
/// \brief общая информация и парметры модуля ядра

#ifndef MIL1553UD_INFO_H
#define MIL1553UD_INFO_H

/// \remark информация о дате и версии
/// \brief старший номер версии
#define DRIVER_MAJOR_VER                2
/// \brief младший номер версии
#define DRIVER_MINOR_VER                4
/// \brief день создания
#define DRIVER_DATE_DAY                 3
/// \brief месяц создания
#define DRIVER_DATE_MONTH               11
/// \brief год создания
#define DRIVER_DATE_YEAR                23
/// \brief закодированная дата и версия
///  31..28 - major_ver; 27..24 - minor_ver; 23..16 - day; 15..8 - month; 7..0 - year;
#define DRIVER_DATE_N_VERSION           ((DRIVER_MAJOR_VER & 0xF) << 28) | ((DRIVER_MINOR_VER & 0xF) << 24) | ((DRIVER_DATE_DAY & 0xFF) << 16) | ((DRIVER_DATE_MONTH & 0xFF) << 8) | (DRIVER_DATE_YEAR & 0xFF)

/// \brief имя драйвера
#define DRIVER_NAME                     "mil1553ud_driver"
/// \brief версия драйвера
#define DRIVER_VERSION                  "2.4"
/// \brief описание драйвера
#define DRIVER_DESCRIPTION              "Mil1553ud pci card driver"
/// \brief права драйвера
#define DRIVER_COPYRIGHT                "'NOVOMAR' LLC"
/// \brief лицензия драйвера
#define DRIVER_LICENSE                  "GPL"
/// \brief шаблон имени pci-устройства
#define PCI_DEVICE_NAME_TEMPLATE        "mil1553ud_dev_"

/// \brief вендор
#define VENDOR_ID_NOVOMAR               0xA203

/// \brief идентификатор поддерживаемого устройства mPCIe-1553UD1
#define DEVICE_ID_MPCIE1553UD1          0x9470
/// \brief идентификатор поддерживаемого устройства mPCIe-1553UD2
#define DEVICE_ID_MPCIE1553UD2          0x9473
/// \brief идентификатор поддерживаемого устройства PCIe(XMC, CPCIS)-1553UDx
#define DEVICE_ID_PCIE1553UDX           0x9472
/// \brief идентификатор поддерживаемого устройства mPCIe-1553UD1T
#define DEVICE_ID_MPCIE1553UD1T         0x9474
/// \brief идентификатор поддерживаемого устройства mPCIe-1553UD2T
#define DEVICE_ID_MPCIE1553UD2T         0x9475
/// \brief идентификатор поддерживаемого устройства PCIe(XMC, CPCIS)-1553UDxT
#define DEVICE_ID_PCIE1553UDXT          0x9476

/// \brief максимально допустимое количество pci-устройств одновременно
#define MAX_COUNT_BOARDS    32
/// \brief максимальное количество каналов на одном pci-устройстве
#define MAX_CHANNELS        4

/// \brief размер буфера ДМА для одного канала
#define DMA_SIZE_CH         1024*1024
/// \brief максимальный размер ДМА буфера
#define DMA_SIZE            DMA_SIZE_CH*4

#endif

