/// \file mil1553ud_ioctl.с
/// \brief инфраструктура для обработки ioctl команд
#include "mil1553ud_ioctl.h"
#include "mil1553ud_log.h"
#include "mil1553ud_regs.h"
#include "mil1553ud_info.h"
#include "mil1553ud_cdev.h"

/// \remark заголовки функций ioctl-команд

long ioctl_write_reg(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_write_reg_bit_mask(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_read_reg(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_switch_mode(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_switch_bus(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_dev_channel_work(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_set_rt_tr_mode(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_set_rt_tr_buf_ready(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_rt_buf_man_en(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_version(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_version_driver(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_pci_location(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_clear_dma(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_enable_dma(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_disable_dma(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_get_nblock_raw_dma(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_read_blocks_raw_dma(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_read_blocks_raw_dma_fix_count(struct mil1553_cdev_data *dev, unsigned long arg);

long ioctl_manage_interrupt(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_man_irq_subaddress_tr(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_man_irq_subaddress_rcv(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_set_address(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_set_timer_rtbm_conf_reg_pci(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_set_timer_bc_conf_reg_pci(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_wr_block_buf_subadr(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_rd_block_buf_subadr(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_write_bc_ram(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_read_bc_ram(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_get_count_interrupt_blocks(struct mil1553_cdev_data* dev, unsigned long arg);
long ioctl_read_interrupt_blocks(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_get_rcv_subadr_data(struct mil1553_cdev_data* dev, unsigned long arg);

//long ioctl_set_signal_irq(struct mil1553_cdev_data* dev, unsigned long arg);
//long ioctl_clr_signal_irq(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_get_number_ch(struct mil1553_cdev_data* dev, unsigned long arg);

long ioctl_handler(struct mil1553_cdev_data* dev, unsigned int cmd, unsigned long arg)
{
    long err = 0;
    if (_IOC_TYPE( cmd ) != IOC_MAGIC) {
        LOG_ERROR("ioctl_handler - bad IOC_MAGIC");
        return -ENOTTY;
    }
    LOG_INFO("devIndex = %d, chNum = %d, drvIndex = %d", dev->index, dev->numCh, dev->drv->unit);
    switch (cmd) {
    // взаимодействие с регистрами
    case IOCTL_WRITE_REG:                       err = ioctl_write_reg(dev, arg);
        break;
    case IOCTL_WRITE_REG_BIT_MASK:              err = ioctl_write_reg_bit_mask(dev, arg);
        break;
    case IOCTL_READ_REG:                        err = ioctl_read_reg(dev, arg);
        break;
    // РЕЖИМЫ
    case IOCTL_SWITCH_MODE:                     err = ioctl_switch_mode(dev, arg);
        break;
    case IOCTL_SWITCH_BUS:                      err = ioctl_switch_bus(dev, arg);
        break;
    case IOCTL_DEV_CHANNEL_WORK:                err = ioctl_dev_channel_work(dev, arg);
        break;
    // term dev
    case IOCTL_SET_RT_TR_MODE:                  err = ioctl_set_rt_tr_mode(dev, arg);
        break;
    case IOCTL_SET_RT_TR_BUF_READY:             err = ioctl_set_rt_tr_buf_ready(dev, arg);
        break;
    case IOCTL_RT_BUF_MAN_EN:                   err = ioctl_rt_buf_man_en(dev, arg);
        break;
    // DMA
    case IOCTL_CLEAR_DMA:                       err = ioctl_clear_dma(dev, arg);
        break;
    case IOCTL_READ_BLOCKS_RAW_DMA:             err = ioctl_read_blocks_raw_dma(dev, arg);
        break;
    case IOCTL_GET_NBLOCK_RAW_DMA:              err = ioctl_get_nblock_raw_dma(dev, arg);
        break;
    case IOCTL_READ_BLOCKS_RAW_DMA_FIX_COUNT:   err = ioctl_read_blocks_raw_dma_fix_count(dev, arg);
        break;
    case IOCTL_ENABLE_DMA:                      err = ioctl_enable_dma(dev, arg);
        break;
    case IOCTL_DISABLE_DMA:                     err = ioctl_disable_dma(dev, arg);
        break;
    // IRQ
    case IOCTL_MANAGE_INTERRUPT:                err = ioctl_manage_interrupt(dev, arg);
        break;
    case IOCTL_MAN_IRQ_SUBADDRESS_TR:           err = ioctl_man_irq_subaddress_tr(dev, arg);
        break;
    case IOCTL_MAN_IRQ_SUBADDRESS_RCV:          err = ioctl_man_irq_subaddress_rcv(dev, arg);
        break;
    // КАНАЛЫ
    case IOCTL_SET_ADDRESS:                     err = ioctl_set_address(dev, arg);
        break;
    case IOCTL_SET_TIMER_RTBM_CONF_REG_PCI:     err = ioctl_set_timer_rtbm_conf_reg_pci(dev, arg);
        break;
    case IOCTL_SET_TIMER_BC_CONF_REG_PCI:       err = ioctl_set_timer_bc_conf_reg_pci(dev, arg);
        break;
    // ПОДАДРЕСА И БУФЕРЫ
    case IOCTL_WR_BLOCK_BUF_SUBADR:             err = ioctl_wr_block_buf_subadr(dev, arg);
        break;
    case IOCTL_RD_BLOCK_BUF_SUBADR:             err = ioctl_rd_block_buf_subadr(dev, arg);
        break;
    // bc_ram_reg
    case IOCTL_WRITE_BC_RAM:                    err = ioctl_write_bc_ram(dev, arg);
        break;
    case IOCTL_READ_BC_RAM:                     err = ioctl_read_bc_ram(dev, arg);
        break;
    // info command
    case IOCTL_VERSION:                         err = ioctl_version(dev, arg);
        break;
    case IOCTL_VERSION_DRIVER:                  err = ioctl_version_driver(dev, arg);
        break;
    case IOCTL_PCI_LOCATION:                    err = ioctl_pci_location(dev, arg);
        break;
    // interrupt ring buffer
    case IOCTL_GET_COUNT_INTERRUPT_BLOCKS:      err = ioctl_get_count_interrupt_blocks(dev, arg);
        break;
    case IOCTL_READ_INTERRUPT_BLOCKS:           err = ioctl_read_interrupt_blocks(dev, arg);
        break;
    case IOCTL_GET_RCV_SUBADR_DATA:             err = ioctl_get_rcv_subadr_data(dev, arg);
        break;
//    case IOCTL_SET_SIGNAL_IRQ:                  err = ioctl_set_signal_irq(dev, arg);
//        break;
//    case IOCTL_CLR_SIGNAL_IRQ:                  err = ioctl_clr_signal_irq(dev, arg);
//        break;
    // сервис
    case IOCTL_GET_NUMBER_CH:                   err = ioctl_get_number_ch(dev, arg);
        break;
    default:
        LOG_ERROR("ioctl_handler - understand cmd - %d", cmd);
        return -ENOTTY;
    }
    // обработка ошибок
    if (err)
        LOG_ERROR("ioctl_handler - error - %d", err);
    return err;
}

/// \remark реализация ioctl-команд

/// \remark константы, необходимые для чтения/записи регистров

#define BORDER_COMMON                  0x2000
/// \remark канал 1
#define BORDER_LO_CH1_START            0x2000
#define BORDER_LO_CH1_FINISH           0x4000
#define BORDER_HI_CH1_START            0x10000
#define BORDER_HI_CH1_FINISH           0x20000
/// \remark канал 2
#define BORDER_LO_CH2_START            0x4000
#define BORDER_LO_CH2_FINISH           0x6000
#define BORDER_HI_CH2_START            0x20000
#define BORDER_HI_CH2_FINISH           0x30000
/// \remark канал 3
#define BORDER_LO_CH3_START            0x6000
#define BORDER_LO_CH3_FINISH           0x8000
#define BORDER_HI_CH3_START            0x30000
#define BORDER_HI_CH3_FINISH           0x40000
/// \remark канал 4
#define BORDER_LO_CH4_START            0x8000
#define BORDER_LO_CH4_FINISH           0x10000
#define BORDER_HI_CH4_START            0x40000
#define BORDER_HI_CH4_FINISH           0x50000

long ioctl_write_reg(struct mil1553_cdev_data* dev, unsigned long arg)
{
    SADDR_DATA data;
    LOG_INFO("ioctl_write_reg");
    copy_from_user((void *) &data, (void *) arg, sizeof(SADDR_DATA));
    if (data.daddr < BORDER_COMMON) { // пишем через мьютекс общий
        mutex_lock(&dev->drv->mtx);
        iowrite32(data.data, dev->drv->pbar + data.daddr);
        mutex_unlock(&dev->drv->mtx);
    } else { // поканальное обращение
        switch (dev->numCh) {
        case CH_NUM_1:
            if (((data.daddr >= BORDER_LO_CH1_START) && (data.daddr < BORDER_LO_CH1_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH1_START) && (data.daddr < BORDER_HI_CH1_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                iowrite32(data.data, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 1");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_2:
            if (((data.daddr >= BORDER_LO_CH2_START) && (data.daddr < BORDER_LO_CH2_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH2_START) && (data.daddr < BORDER_HI_CH2_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                iowrite32(data.data, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 2");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_3:
            if (((data.daddr >= BORDER_LO_CH3_START) && (data.daddr < BORDER_LO_CH3_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH3_START) && (data.daddr < BORDER_HI_CH3_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                iowrite32(data.data, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 3");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_4:
            if (((data.daddr >= BORDER_LO_CH4_START) && (data.daddr < BORDER_LO_CH4_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH4_START) && (data.daddr < BORDER_HI_CH4_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                iowrite32(data.data, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 4");
                return ERR_BAD_ADDRESS;
            }
            break;
        default:
            LOG_ERROR("bad channel");
            return ERR_BAD_CHANNEL;
        }
    }
    return GOOD;
}

/// \brief макрос записи заданной битовой маски в регистр
#define SET_REG_BIT_MASK(dev, data, regVal, shift)  \
    regVal = ioread32(dev->drv->pbar + data.daddr); /* считываем значение регистра */ \
    for(shift = 0; shift < 32; shift++) { /* обходим все биты */ \
        if (data.mask & (1 << shift)) { /* 1 - изменяем бит */ \
            if (data.data & (1 << shift)) /* смотрим исходное значение бита */ \
                regVal |= (1 << shift); /* устанавливаем 1 */ \
            else \
                regVal &= ~(1 << shift); /* устанавливаем 0 */ \
        } else { /* 0 - не изменяем бит */ \
            continue; \
        } \
    }

long ioctl_write_reg_bit_mask(struct mil1553_cdev_data* dev, unsigned long arg)
{
    SADDR_DATA_BIT_MASK data;   // исходные данные
    unsigned int regVal;        // временное значение регистра
    int shift;                  // смещение бита
    LOG_INFO("ioctl_write_reg_bit_mask");
    copy_from_user((void *) &data, (void *) arg, sizeof(SADDR_DATA_BIT_MASK));
    if (data.daddr < BORDER_COMMON) { // пишем через мьютекс общий
        mutex_lock(&dev->drv->mtx);
        SET_REG_BIT_MASK(dev, data, regVal, shift);
        iowrite32(regVal, dev->drv->pbar + data.daddr);
        mutex_unlock(&dev->drv->mtx);
    } else { // поканальное обращение
        switch (dev->numCh) {
        case CH_NUM_1:
            if (((data.daddr >= BORDER_LO_CH1_START) && (data.daddr < BORDER_LO_CH1_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH1_START) && (data.daddr < BORDER_HI_CH1_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                regVal = ioread32(dev->drv->pbar + data.daddr); /* считываем значение регистра */
                for(shift = 0; shift < 32; shift++) { /* обходим все биты */
                    if (data.mask & (1 << shift)) { /* 1 - изменяем бит */
                        if (data.data & (1 << shift)) /* смотрим исходное значение бита */
                            regVal |= (1 << shift); /* устанавливаем 1 */
                        else \
                            regVal &= ~(1 << shift); /* устанавливаем 0 */
                    } else { /* 0 - не изменяем бит */
                        continue;
                    }
                }
                iowrite32(regVal, dev->drv->pbar + data.daddr);
                LOG_INFO("adr = 0x%x, val = 0x%x", data.daddr, regVal);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 1");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_2:
            if (((data.daddr >= BORDER_LO_CH2_START) && (data.daddr < BORDER_LO_CH2_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH2_START) && (data.daddr < BORDER_HI_CH2_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                SET_REG_BIT_MASK(dev, data, regVal, shift);
                iowrite32(regVal, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 2");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_3:
            if (((data.daddr >= BORDER_LO_CH3_START) && (data.daddr < BORDER_LO_CH3_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH3_START) && (data.daddr < BORDER_HI_CH3_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                SET_REG_BIT_MASK(dev, data, regVal, shift);
                iowrite32(regVal, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 3");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_4:
            if (((data.daddr >= BORDER_LO_CH4_START) && (data.daddr < BORDER_LO_CH4_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH4_START) && (data.daddr < BORDER_HI_CH4_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                SET_REG_BIT_MASK(dev, data, regVal, shift);
                iowrite32(regVal, dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 4");
                return ERR_BAD_ADDRESS;
            }
            break;
        default:
            LOG_ERROR("bad channel");
            return ERR_BAD_CHANNEL;
        }
    }
    return GOOD;
}

long ioctl_read_reg(struct mil1553_cdev_data* dev, unsigned long arg)
{
    SADDR_DATA data;
    LOG_INFO("ioctl_read_reg");
    copy_from_user((void *) &data, (void *) arg, sizeof(SADDR_DATA));
    if (data.daddr < BORDER_COMMON) { // пишем через мьютекс общий
        mutex_lock(&dev->drv->mtx);
        data.data = ioread32(dev->drv->pbar + data.daddr);
        mutex_unlock(&dev->drv->mtx);
    } else { // поканальное обращение
        switch (dev->numCh) {
        case CH_NUM_1:
            if (((data.daddr >= BORDER_LO_CH1_START) && (data.daddr < BORDER_LO_CH1_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH1_START) && (data.daddr < BORDER_HI_CH1_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                data.data = ioread32(dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 1");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_2:
            if (((data.daddr >= BORDER_LO_CH2_START) && (data.daddr < BORDER_LO_CH2_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH2_START) && (data.daddr < BORDER_HI_CH2_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                data.data = ioread32(dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 2");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_3:
            if (((data.daddr >= BORDER_LO_CH3_START) && (data.daddr < BORDER_LO_CH3_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH3_START) && (data.daddr < BORDER_HI_CH3_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                data.data = ioread32(dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 3");
                return ERR_BAD_ADDRESS;
            }
            break;
        case CH_NUM_4:
            if (((data.daddr >= BORDER_LO_CH4_START) && (data.daddr < BORDER_LO_CH4_FINISH)) ||
                ((data.daddr >= BORDER_HI_CH4_START) && (data.daddr < BORDER_HI_CH4_FINISH)) ) {
                mutex_lock(&dev->mtxCh);
                data.data = ioread32(dev->drv->pbar + data.daddr);
                mutex_unlock(&dev->mtxCh);
            } else {
                LOG_ERROR("bad addr for channel 4");
                return ERR_BAD_ADDRESS;
            }
            break;
        default:
            LOG_ERROR("bad channel");
            return ERR_BAD_CHANNEL;
        }
    }
    if (copy_to_user((void*)arg, (void*)&data, sizeof(SADDR_DATA)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_switch_mode(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned int mode;          // режим работы канала
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_switch_mode");
    copy_from_user((void *) &mode, (void *) arg, sizeof(unsigned int));
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, CTRL_REG_PCI_CH1, CTRL_REG_PCI_CH2, CTRL_REG_PCI_CH3, CTRL_REG_PCI_CH4);
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal &= ~(0x3);
        switch (mode) {
        case MIL_MODE_BUS_CONTR:         regVal |= MIL_MODE_BUS_CONTR;
            break;
        case MIL_MODE_MONITOR:           regVal |= MIL_MODE_MONITOR;
            break;
        case MIL_MODE_MONITOR_ADRR:      regVal |= MIL_MODE_MONITOR_ADRR;
            break;
        case MIL_MODE_TERMINAL_DEV:      regVal |= MIL_MODE_TERMINAL_DEV;
            break;
        default:
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument");
            return ERR_BAD_ARG_MODE;
        }
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_switch_bus(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned int mode;          // режим работы канала
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_switch_bus");
    copy_from_user((void *) &mode, (void *) arg, sizeof(unsigned int));
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, CTRL_REG_PCI_CH1, CTRL_REG_PCI_CH2, CTRL_REG_PCI_CH3, CTRL_REG_PCI_CH4);
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal &= ~(0xC);
        if (mode & MIL_BUS_A_EN)
            regVal |= (1 << 2);
        if (mode & MIL_BUS_B_EN)
            regVal |= (1 << 3);
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_dev_channel_work(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned int mode;          // режим работы канала
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_dev_channel_work");
    copy_from_user((void *) &mode, (void *) arg, sizeof(unsigned int));
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, CTRL_REG_PCI_CH1, CTRL_REG_PCI_CH2, CTRL_REG_PCI_CH3, CTRL_REG_PCI_CH4);
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal &= ~(1 << 23);
        switch (mode) {
        case MIL_DEV_CHANNEL_ON:          regVal |= (1 << 23);
            break;
        case MIL_DEV_CHANNEL_OFF:         regVal &= ~(1 << 23);
            break;
        default:
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument");
            return ERR_BAD_ARG_MODE;
        }
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_set_rt_tr_mode(struct mil1553_cdev_data* dev, unsigned long arg)
{
    RT_TR_MODE mode;            // режим работы ОУ
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_set_rt_tr_mode");
    copy_from_user((void *) &mode, (void *) arg, sizeof(RT_TR_MODE));
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_TR_REG_CH1, RT_TR_REG_CH2, RT_TR_REG_CH3, RT_TR_REG_CH4);
        if ((mode.subaddress >= 1) && (mode.subaddress <= 30)) {
            regAddr += 0x4 * mode.subaddress;   // выбираем заданный подадрес
        } else {
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument - subaddress");
            return ERR_BAD_ARG_SA;
        }
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal |= (1 << 31);    // разрешаем писать режим
        regVal &= ~((1 << 28) | (1 << 29)); // обнуляем режим
        switch (mode.mode) {
        case MIL_RT_MODE_PROG:
            iowrite32(regVal, dev->drv->pbar + regAddr);
            break;
        case MIL_RT_MODE_HRDW:
            regVal |= 1 << 28;
            iowrite32(regVal, dev->drv->pbar + regAddr);
            break;
        default:
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument - mode");
            return ERR_BAD_ARG_MODE;
        }
        regVal &= ~(1 << 31);    // сохраняем записанное
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_set_rt_tr_buf_ready(struct mil1553_cdev_data* dev, unsigned long arg)
{
    RT_TR_MODE_BUF modeBuf;     // режим работы ОУ буфера
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_set_rt_tr_buf_ready");
    copy_from_user((void *) &modeBuf, (void *) arg, sizeof(RT_TR_MODE_BUF));
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_TR_REG_CH1, RT_TR_REG_CH2, RT_TR_REG_CH3, RT_TR_REG_CH4);
        if ((modeBuf.subaddress >= 1) && (modeBuf.subaddress <= 30)) {
            regAddr += 0x4 * modeBuf.subaddress;   // выбираем заданный подадрес
        } else {
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument - subaddress");
            return ERR_BAD_ARG_SA;
        }
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal |= (1 << 31);    // разрешаем писать режим
        switch (modeBuf.cmd_buf) {
        case MIL_RT_BUF_0_OFF_1_OFF:
            regVal &= ~(1 << 22);   // RTF_BUF0
            regVal &= ~(1 << 23);   // RTF_BUF1
            break;
        case MIL_RT_BUF_0_OFF_1_ON:
            regVal &= ~(1 << 22);   // RTF_BUF0
            regVal |=  (1 << 23);   // RTF_BUF1
            break;
        case MIL_RT_BUF_0_ON_1_OFF:
            regVal |=  (1 << 22);   // RTF_BUF0
            regVal &= ~(1 << 23);   // RTF_BUF1
            break;
        case MIL_RT_BUF_0_ON_1_ON:
            regVal |=  (1 << 22);   // RTF_BUF0
            regVal |=  (1 << 23);   // RTF_BUF1
            break;
        default:
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument - mode");
            return ERR_BAD_ARG_MODE;
        }
        iowrite32(regVal, dev->drv->pbar + regAddr);
        regVal &= ~(1 << 31);    // сохраняем записанное
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_rt_buf_man_en(struct mil1553_cdev_data* dev, unsigned long arg)
{
    RT_BUF_MAN_EN modeBufMan;   // разрешение буферов ОУ
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    int shiftSa;                // смещение бита поддареса
    LOG_INFO("ioctl_rt_buf_man_en");
    copy_from_user((void *) &modeBufMan, (void *) arg, sizeof(RT_BUF_MAN_EN));
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        switch (modeBufMan.direction) { // выбор направления
        case MIL_RT_BUF_TRANSMIT:
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_TR_REG_CH1, RT_TR_REG_CH2, RT_TR_REG_CH3, RT_TR_REG_CH4);
            for(shiftSa = 1; shiftSa <= 30; shiftSa++) {
                if (modeBufMan.subaddress_mask & (1 << shiftSa)) {
                    regVal = ioread32(dev->drv->pbar + regAddr + 0x4 * shiftSa);
                    regVal |= (1 << 31);    // разрешаем писать режим
                    iowrite32(regVal, dev->drv->pbar + regAddr + 0x4 * shiftSa);
                    if (modeBufMan.action_mask & (1 << shiftSa)) {
                        regVal |= 1 << 30;  // разрешаем буфер поддареса shiftSa
                    } else {
                        regVal &= ~(1 << 30); // запрещаем буфер поддареса shiftSa
                    }
                    iowrite32(regVal, dev->drv->pbar + regAddr + 0x4 * shiftSa);
                    regVal &= ~(1 << 31);    // сохраняем записанное
                    iowrite32(regVal, dev->drv->pbar + regAddr + 0x4 * shiftSa);
                } else
                    continue;
            }
            break;
        case MIL_RT_BUF_RECEIVE:
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_RCV_REG_CH1, RT_RCV_REG_CH2, RT_RCV_REG_CH3, RT_RCV_REG_CH4);
            for(shiftSa = 1; shiftSa <= 30; shiftSa++) {
                if (modeBufMan.subaddress_mask & (1 << shiftSa)) {
                    regVal = ioread32(dev->drv->pbar + regAddr + 0x4 * shiftSa);
                    if (modeBufMan.action_mask & (1 << shiftSa)) {
                        regVal &= ~(1 << 31);  // разрешаем буфер поддареса shiftSa
                    } else {
                        regVal |= (1 << 31);  // запрещаем буфер поддареса shiftSa
                    }
                    iowrite32(regVal, dev->drv->pbar + regAddr + 0x4 * shiftSa);
                } else
                    continue;
            }
            break;
        default:
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument - direction");
            return ERR_BAD_ARG_DIRECTION;
        }
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_version(struct mil1553_cdev_data* dev, unsigned long arg)
{
    int i;              // счётчик
    VERSION version;    // подробная информация о драйвере и устройстве
    LOG_INFO("ioctl_version");
    version.vendor_id = dev->drv->vendorid;
    version.device_id = dev->drv->deviceid;
    version.revision = dev->drv->revisionid;
    version.type = dev->drv->current_count_of_channels;
    for (i = 0; i < 30; i++)
        version.dev_name[i] = dev->drv->devname[i];
    version.minor = dev->numCh;
    version.irq = dev->drv->pdevice->irq;
    version.size_dma = dev->drv->dma_info.dma_size;
    version.addr_dma_virt = dev->drv->dma_info.dma_vaddr;
    version.pci_bars = dev->drv->bar;
    version.addr_bar_virt = dev->drv->pbar;
    if (copy_to_user((void*)arg, (void*)&version, sizeof(VERSION)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_version_driver(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned int driverVersion;     // версия драйвера
    LOG_INFO("ioctl_version_driver");
    driverVersion = DRIVER_DATE_N_VERSION;
    if (copy_to_user((void*)arg, (void*)&driverVersion, sizeof(unsigned int)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_pci_location(struct mil1553_cdev_data* dev, unsigned long arg)
{
    int i;                      // счётчик
    PCI_LOCATION pciLoc;        // информация о pci-локации устройства
    LOG_INFO("ioctl_pci_location");
    for ( i = 0; i < SIZE_OF_PCI_LOC; i++)
        pciLoc.name[i] = 0;
    snprintf(pciLoc.name, SIZE_OF_PCI_LOC, "%s." , pci_name(dev->drv->pdevice));
    if (copy_to_user((void*)arg, (void*)&pciLoc, sizeof(PCI_LOCATION)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_clear_dma(struct mil1553_cdev_data* dev, unsigned long arg)
{
   LOG_INFO("ioctl_clear_dma");
   switch (dev->numCh) {
   case CH_NUM_1:
       mutex_lock(&dev->mtxCh);
       iowrite32(0, dev->drv->pbar + DMA_INDEX_CH1);
       dev->drv->dma_info.ch_dma_write[CH_NUM_1] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + CH_NUM_1 * DMA_SIZE_CH);
       dev->drv->dma_info.ch_dma_read[CH_NUM_1] = (uint32_t*)(dev->drv->dma_info.ch_dma_write[CH_NUM_1]);
       dev->interrupt_buffer->index_read = 0;
       dev->interrupt_buffer->index_write = 0;
       mutex_unlock(&dev->mtxCh);
       break;
   case CH_NUM_2:
       mutex_lock(&dev->mtxCh);
       iowrite32(0, dev->drv->pbar + DMA_INDEX_CH2);
       dev->drv->dma_info.ch_dma_write[CH_NUM_2] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + CH_NUM_2 * DMA_SIZE_CH);
       dev->drv->dma_info.ch_dma_read[CH_NUM_2] = (uint32_t*)(dev->drv->dma_info.ch_dma_write[CH_NUM_2]);
       dev->interrupt_buffer->index_read = 0;
       dev->interrupt_buffer->index_write = 0;
       mutex_unlock(&dev->mtxCh);
       break;
   case CH_NUM_3:
       mutex_lock(&dev->mtxCh);
       iowrite32(0, dev->drv->pbar + DMA_INDEX_CH3);
       dev->drv->dma_info.ch_dma_write[CH_NUM_3] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + CH_NUM_3 * DMA_SIZE_CH);
       dev->drv->dma_info.ch_dma_read[CH_NUM_3] = (uint32_t*)(dev->drv->dma_info.ch_dma_write[CH_NUM_3]);
       dev->interrupt_buffer->index_read = 0;
       dev->interrupt_buffer->index_write = 0;
       mutex_unlock(&dev->mtxCh);
       break;
   case CH_NUM_4:
       mutex_lock(&dev->mtxCh);
       iowrite32(0, dev->drv->pbar + DMA_INDEX_CH4);
       dev->drv->dma_info.ch_dma_write[CH_NUM_4] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + CH_NUM_4 * DMA_SIZE_CH);
       dev->drv->dma_info.ch_dma_read[CH_NUM_4] = (uint32_t*)(dev->drv->dma_info.ch_dma_write[CH_NUM_4]);
       dev->interrupt_buffer->index_read = 0;
       dev->interrupt_buffer->index_write = 0;
       mutex_unlock(&dev->mtxCh);
       break;
   default:
       LOG_ERROR("bad channel");
       return ERR_BAD_CHANNEL;
   }
   return GOOD;
}

long ioctl_enable_dma(struct mil1553_cdev_data* dev, unsigned long arg)
{
    uint32_t value;
    LOG_INFO("ioctl_enable_dma");
    mutex_lock(&dev->drv->mtx);
    value = ioread32(dev->drv->pbar + DMA_DATA_BASE);
    value |= 0x1;
    iowrite32(value, dev->drv->pbar + DMA_DATA_BASE);
    mutex_unlock(&dev->drv->mtx);
    return GOOD;
}

long ioctl_disable_dma(struct mil1553_cdev_data* dev, unsigned long arg)
{
    uint32_t value;
    LOG_INFO("ioctl_disable_dma");
    mutex_lock(&dev->drv->mtx);
    value = ioread32(dev->drv->pbar + DMA_DATA_BASE);
    value &= ~(0x1);
    iowrite32(value, dev->drv->pbar + DMA_DATA_BASE);
    mutex_unlock(&dev->drv->mtx);
    return GOOD;
}

/// \brief значение адреса указателя записи для канала
#define UINT32_DMA_WRITE(num_ch)            (uint64_t)(dev->drv->dma_info.ch_dma_write[num_ch])
/// \brief значение адреса указателя чтения для канала
#define UINT32_DMA_READ(num_ch)             (uint64_t)(dev->drv->dma_info.ch_dma_read[num_ch])
/// \brief значение базового адреса указателя буфера ДМА
#define UINT32_DMA_VADR                     (uint64_t)(dev->drv->dma_info.dma_vaddr)
/// \brief чтение дма индекса
#define GET_DMA_INDEX(reg)                  \
    (ioread32(dev->drv->pbar + reg) >> 7)
/// \brief вычисления указателя чтения
#define CALC_DMA_WRITE_CH(num_ch)           \
    dev->drv->dma_info.ch_dma_write[num_ch] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + num_ch * DMA_SIZE_CH + dma_index * SIZE_RAW_DMA_BLOCK)
/// \brief вычисление доступных блоков для чтения
#define CALC_NBLOCK_CH(nblock, num_ch)      \
    if (UINT32_DMA_WRITE(num_ch) >= UINT32_DMA_READ(num_ch)) { \
        nblock = (UINT32_DMA_WRITE(num_ch) - UINT32_DMA_READ(num_ch)) / SIZE_RAW_DMA_BLOCK; \
    } else { \
        nblock = (UINT32_DMA_READ(num_ch) - (UINT32_DMA_VADR + num_ch * DMA_SIZE_CH)) / SIZE_RAW_DMA_BLOCK + \
        ((UINT32_DMA_VADR + num_ch * DMA_SIZE_CH + MEGABYTE) - UINT32_DMA_WRITE(num_ch)) / SIZE_RAW_DMA_BLOCK; \
    }

#define BYTES_IN_MEGABYTE 1048567
#define DMABLOCKS_IN_MEGABYTE   BYTES_IN_MEGABYTE / SIZE_RAW_DMA_BLOCK

long ioctl_get_nblock_raw_dma(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned int regAddr;
    unsigned int nblock; // кол-во блоков, доступных для чтения
    uint32_t dma_index; // индекс дма из регистра DMA_INDEX
    LOG_INFO("ioctl_get_nblock_raw_dma");
    mutex_lock(&dev->mtxCh);
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, DMA_INDEX_CH1, DMA_INDEX_CH2, DMA_INDEX_CH3, DMA_INDEX_CH4);
        dma_index = GET_DMA_INDEX(regAddr);
        if (dma_index > DMABLOCKS_IN_MEGABYTE) {
            LOG_ERROR("bad dma_index= %d", dma_index);
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_DMA_INDEX;
        }
    } else {
        LOG_ERROR("bad channel");
        mutex_unlock(&dev->mtxCh);
        return ERR_BAD_CHANNEL;
    }
    dev->drv->dma_info.ch_dma_write[dev->numCh] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH + dma_index * SIZE_RAW_DMA_BLOCK);
    //CALC_NBLOCK_CH(nblock, dev->numCh); // высчитываем доступные для чтения блоки
    if ((uint64_t)dev->drv->dma_info.ch_dma_write[dev->numCh] >= (uint64_t)dev->drv->dma_info.ch_dma_read[dev->numCh]) {
        nblock = (((uint64_t)dev->drv->dma_info.ch_dma_write[dev->numCh] - (uint64_t)dev->drv->dma_info.ch_dma_read[dev->numCh])) / SIZE_RAW_DMA_BLOCK;
        if (nblock > DMABLOCKS_IN_MEGABYTE){
            LOG_ERROR("bad nblock (if) = %d", nblock);
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_DMA_NBLOCK;
        }
    } else {
        nblock = (UINT32_DMA_WRITE(dev->numCh) - (UINT32_DMA_VADR + dev->numCh * DMA_SIZE_CH)) / SIZE_RAW_DMA_BLOCK + \
                ((UINT32_DMA_VADR + dev->numCh * DMA_SIZE_CH + MEGABYTE) - UINT32_DMA_READ(dev->numCh)) / SIZE_RAW_DMA_BLOCK;
        if (nblock > DMABLOCKS_IN_MEGABYTE){
            LOG_ERROR("bad nblock (else) = %d", nblock);
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_DMA_NBLOCK;
        }
    }
    mutex_unlock(&dev->mtxCh);
    if (copy_to_user((void*)arg, (void*)&nblock, sizeof(unsigned int)))
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

/// \brief значение адреса конца буфера дма для канала
#define UINT32_DMA_CH_END(num_ch)  UINT32_DMA_VADR + num_ch * DMA_SIZE_CH + DMA_SIZE_CH

long ioctl_read_blocks_raw_dma(struct mil1553_cdev_data* dev, unsigned long arg)
{
    char isError;               // ошибка
    unsigned int nblock;        // кол-во блоков, доступных для чтения
    unsigned int nblockend;     // кол-во блоков, до границы буфера
    uint32_t dma_index;         // индекс дма из регистра DMA_INDEX
    DMA_READ_BLOCK block;       // данные из юзерспейс
    LOG_INFO("ioctl_read_blocks_raw_dma");
    copy_from_user((void *) &block, (void *) arg, sizeof(DMA_READ_BLOCK)); // получаем данные из юзерспейс
    mutex_lock(&dev->mtxCh);
    switch (dev->numCh) { // вычисляем текущий дма индекс
    case CH_NUM_1:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH1);
        break;
    case CH_NUM_2:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH2);
        break;
    case CH_NUM_3:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH3);
        break;
    case CH_NUM_4:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH4);
        break;
    default:
        LOG_ERROR("bad channel");
        mutex_unlock(&dev->mtxCh);
        return ERR_BAD_CHANNEL;
    }
    CALC_DMA_WRITE_CH(dev->numCh);  // высчитываем указатель записи
    CALC_NBLOCK_CH(nblock, dev->numCh); // высчитываем доступные для чтения блоки
    if (block.countBlocks > nblock)
        block.countBlocks = nblock;
    nblockend = (UINT32_DMA_CH_END(dev->numCh) - UINT32_DMA_READ(dev->numCh)) / SIZE_RAW_DMA_BLOCK; // вычисляем кол-во блоков, до границы буфера
    if (block.countBlocks <= nblockend) { // кол-во запрошенных блоков не превышает границу буфера
        isError = copy_to_user((void*)block.data, (void*)dev->drv->dma_info.ch_dma_read[dev->numCh], block.countBlocks * SIZE_RAW_DMA_BLOCK); // копируем в юзерспейс блоки
        dev->drv->dma_info.ch_dma_read[dev->numCh] += block.countBlocks * SIZE_RAW_DMA_BLOCK / 4; // сдвигаем указатель чтения
        if (UINT32_DMA_READ(dev->numCh) > UINT32_DMA_CH_END(dev->numCh))
            dev->drv->dma_info.ch_dma_read[dev->numCh] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH); // при необходимости корректируем
    } else {
        isError = copy_to_user((void*)block.data, (void*)dev->drv->dma_info.ch_dma_read[dev->numCh], nblockend * SIZE_RAW_DMA_BLOCK);  // копируем первый кусок блоков
        dev->drv->dma_info.ch_dma_read[dev->numCh] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH);
        isError = copy_to_user((void*)(block.data + nblockend * SIZE_RAW_DMA_BLOCK), (void*)dev->drv->dma_info.ch_dma_read[dev->numCh], (block.countBlocks - nblockend) * SIZE_RAW_DMA_BLOCK); // копируем второй кусок блоков
        dev->drv->dma_info.ch_dma_read[dev->numCh] = (uint32_t*)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH + (block.countBlocks - nblockend) * SIZE_RAW_DMA_BLOCK);
    }
    copy_to_user((void*)arg, (void*)&block.countBlocks, sizeof(unsigned int));  // копируем в юзерспей фактическое кол-во блоков полученных
    mutex_unlock(&dev->mtxCh);
    if (isError)
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

long ioctl_read_blocks_raw_dma_fix_count(struct mil1553_cdev_data *dev, unsigned long arg)
{
    char isError;           // ошибка
    unsigned int nblock;    // кол-во блоков, доступных для чтения
    unsigned int nblockend; // кол-во блоков, до границы буфера
    uint32_t dma_index;     // индекс дма из регистра DMA_INDEX
    MIL1553_DMA_FIX_BLOCK block;   // данные из юзерспейс
    LOG_INFO("ioctl_read_blocks_raw_dma_fix_count");
    copy_from_user((void *)&block, (void *)arg, sizeof(MIL1553_DMA_FIX_BLOCK)); // получаем данные из юзерспейс
    mutex_lock(&dev->mtxCh);
    switch (dev->numCh)
    { // вычисляем текущий дма индекс
    case CH_NUM_1:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH1);
        break;
    case CH_NUM_2:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH2);
        break;
    case CH_NUM_3:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH3);
        break;
    case CH_NUM_4:
        dma_index = GET_DMA_INDEX(DMA_INDEX_CH4);
        break;
    default:
        LOG_ERROR("bad channel");
        mutex_unlock(&dev->mtxCh);
        return ERR_BAD_CHANNEL;
    }
    block.currentBlocks = 0;
    CALC_DMA_WRITE_CH(dev->numCh);      // высчитываем указатель записи
    CALC_NBLOCK_CH(nblock, dev->numCh); // высчитываем доступные для чтения блоки

    if (nblock > block.mallocedBlocks)
        block.currentBlocks = block.mallocedBlocks;
    else
        block.currentBlocks = nblock;
    block.nextBlocks = nblock - block.currentBlocks;

    nblockend = (UINT32_DMA_CH_END(dev->numCh) - UINT32_DMA_READ(dev->numCh)) / MIL1553_ONE_BLOCK_SIZE_128_BYTES; // вычисляем кол-во блоков, до границы буфера
    if (block.currentBlocks <= nblockend) { // кол-во запрошенных блоков не превышает границу буфера
        isError = copy_to_user(
                    (void *)block.data, /*куда*/
                    (void *)dev->drv->dma_info.ch_dma_read[dev->numCh], /*откуда*/
                    block.currentBlocks * MIL1553_ONE_BLOCK_SIZE_128_BYTES /*сколько байт*/
                );
        dev->drv->dma_info.ch_dma_read[dev->numCh] += block.currentBlocks * MIL1553_ONE_BLOCK_SIZE_128_BYTES / 4;                                               // сдвигаем указатель чтения
        if (UINT32_DMA_READ(dev->numCh) > UINT32_DMA_CH_END(dev->numCh))
            dev->drv->dma_info.ch_dma_read[dev->numCh] = (uint32_t *)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH); // при необходимости корректируем
    } else {
        isError = copy_to_user(
                    (void *)block.data, /*куда*/
                    (void *)dev->drv->dma_info.ch_dma_read[dev->numCh], /*откуда*/
                    nblockend * MIL1553_ONE_BLOCK_SIZE_128_BYTES /*сколько байт*/
                );
        dev->drv->dma_info.ch_dma_read[dev->numCh] = (uint32_t *)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH);
        // копируем второй кусок блоков
        isError = copy_to_user(
                    (void *)(block.data + nblockend * MIL1553_ONE_BLOCK_SIZE_128_BYTES), /*куда*/
                    (void *)dev->drv->dma_info.ch_dma_read[dev->numCh], /*откуда*/
                    (block.currentBlocks - nblockend) * MIL1553_ONE_BLOCK_SIZE_128_BYTES /*сколько байт*/
                );

        dev->drv->dma_info.ch_dma_read[dev->numCh] = (uint32_t *)(dev->drv->dma_info.dma_vaddr + dev->numCh * DMA_SIZE_CH + (block.currentBlocks - nblockend) * MIL1553_ONE_BLOCK_SIZE_128_BYTES);
    }
    copy_to_user((void *)arg, (void *)&block, sizeof(MIL1553_DMA_FIX_BLOCK)); // копируем в юзерспей
    mutex_unlock(&dev->mtxCh);
    if (isError)
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

/// \brief управление битами прерывания
#define INTERRUPT_MANAGE(intState, shift, regVal)   \
    switch (intState) { \
    case INTERRUPT_MAN_ON:      regVal |= 1<<shift;    break; \
    case INTERRUPT_MAN_OFF:     regVal &= ~(1<<shift); break; \
    }

long ioctl_manage_interrupt(struct mil1553_cdev_data* dev, unsigned long arg)
{
    INTERRUPT_MAN intMan;       // входные данные по управлению прерыванием
    unsigned int intRegMask;    // регистр INTERRUPT_MASK
    LOG_INFO("ioctl_manage_interrupt");
    copy_from_user((void *) &intMan, (void *) arg, sizeof(INTERRUPT_MAN));
    mutex_lock(&dev->drv->mtx);
    intRegMask = ioread32(dev->drv->pbar + INTERRUPT_MASK);
    // маскируем соответствующие прерывания
    INTERRUPT_MANAGE(intMan.int_flash, 24, intRegMask); // INT_FLASH
    switch (dev->numCh) {
    case CH_NUM_1: {
        INTERRUPT_MANAGE(intMan.int_bc,            10, intRegMask); // INT_BC
        INTERRUPT_MANAGE(intMan.int_rt_ren,         9, intRegMask); // INT_RT_REN
        INTERRUPT_MANAGE(intMan.int_rt_ten,         8, intRegMask); // INT_RT_TEN
        INTERRUPT_MANAGE(intMan.int_qdat,           1, intRegMask); // INT_QDAT
        INTERRUPT_MANAGE(intMan.int_hdat,           0, intRegMask); // INT_HDAT
        INTERRUPT_MANAGE(intMan.int_data_cnt_en,   11, intRegMask); // INT_DATA_CNT_EN
        INTERRUPT_MANAGE(intMan.int_timeout_itven, 24, intRegMask); // INT_TIMEOUT_ITVEN
        INTERRUPT_MANAGE(intMan.int_timeout_absen, 28, intRegMask); // INT_TIMEOUT_ABSEN
    } break;
    case CH_NUM_2: {
        INTERRUPT_MANAGE(intMan.int_bc,            14, intRegMask); // INT_BC
        INTERRUPT_MANAGE(intMan.int_rt_ren,        13, intRegMask); // INT_RT_REN
        INTERRUPT_MANAGE(intMan.int_rt_ten,        12, intRegMask); // INT_RT_TEN
        INTERRUPT_MANAGE(intMan.int_qdat,           3, intRegMask); // INT_QDAT
        INTERRUPT_MANAGE(intMan.int_hdat,           2, intRegMask); // INT_HDAT
        INTERRUPT_MANAGE(intMan.int_data_cnt_en,   15, intRegMask); // INT_DATA_CNT_EN
        INTERRUPT_MANAGE(intMan.int_timeout_itven, 25, intRegMask); // INT_TIMEOUT_ITVEN
        INTERRUPT_MANAGE(intMan.int_timeout_absen, 29, intRegMask); // INT_TIMEOUT_ABSEN
    } break;
    case CH_NUM_3: {
        INTERRUPT_MANAGE(intMan.int_bc,            18, intRegMask); // INT_BC
        INTERRUPT_MANAGE(intMan.int_rt_ren,        17, intRegMask); // INT_RT_REN
        INTERRUPT_MANAGE(intMan.int_rt_ten,        16, intRegMask); // INT_RT_TEN
        INTERRUPT_MANAGE(intMan.int_qdat,           5, intRegMask); // INT_QDAT
        INTERRUPT_MANAGE(intMan.int_hdat,           4, intRegMask); // INT_HDAT
        INTERRUPT_MANAGE(intMan.int_data_cnt_en,   19, intRegMask); // INT_DATA_CNT_EN
        INTERRUPT_MANAGE(intMan.int_timeout_itven, 26, intRegMask); // INT_TIMEOUT_ITVEN
        INTERRUPT_MANAGE(intMan.int_timeout_absen, 30, intRegMask); // INT_TIMEOUT_ABSEN
    } break;
    case CH_NUM_4: {
        INTERRUPT_MANAGE(intMan.int_bc,            22, intRegMask); // INT_BC
        INTERRUPT_MANAGE(intMan.int_rt_ren,        21, intRegMask); // INT_RT_REN
        INTERRUPT_MANAGE(intMan.int_rt_ten,        20, intRegMask); // INT_RT_TEN
        INTERRUPT_MANAGE(intMan.int_qdat,           7, intRegMask); // INT_QDAT
        INTERRUPT_MANAGE(intMan.int_hdat,           6, intRegMask); // INT_HDAT
        INTERRUPT_MANAGE(intMan.int_data_cnt_en,   23, intRegMask); // INT_DATA_CNT_EN
        INTERRUPT_MANAGE(intMan.int_timeout_itven, 27, intRegMask); // INT_TIMEOUT_ITVEN
        INTERRUPT_MANAGE(intMan.int_timeout_absen, 31, intRegMask); // INT_TIMEOUT_ABSEN
    } break;
    default:
        LOG_ERROR("bad channel");
        mutex_unlock(&dev->drv->mtx);
        return ERR_BAD_CHANNEL;
    }
    iowrite32(intRegMask, dev->drv->pbar + INTERRUPT_MASK);
    mutex_unlock(&dev->drv->mtx);
    return GOOD;
}

/// \brief управление битами прерывания по подадресам поканально
#define SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, rtIntRegAdr, nBitInt, rtReg, rtRegAdr, nBitRt, sa)   \
    rtIntReg = ioread32(dev->drv->pbar + rtIntRegAdr); \
    if (irq_sa.rt_int == INTERRUPT_MAN_ON) { rtIntReg |= 1 << nBitInt; } \
    else  if (irq_sa.rt_int == INTERRUPT_MAN_OFF) { rtIntReg &= ~(1 << nBitInt); } \
    iowrite32(rtIntReg, dev->drv->pbar + rtIntRegAdr); \
    for ( sa = 1; sa <= 30; sa++ ) { \
        rtReg = ioread32(dev->drv->pbar + rtRegAdr + sa * 0x4); \
        if (irq_sa.subaddress_mask & (1<<sa)) \
            rtReg |= 1 << nBitRt; \
        else \
            rtReg &= ~(1 << nBitRt); \
        iowrite32(rtReg, dev->drv->pbar + rtRegAdr + sa * 0x4); \
    }

long ioctl_man_irq_subaddress_tr(struct mil1553_cdev_data* dev, unsigned long arg)
{
    INTERRUPT_SUBADDRESS irq_sa;    // входные данные по управлению прерыванием
    unsigned int rtIntReg;          // регистр RT_INT_REG
    unsigned int rtReg;             // регистр RT_TR_REG
    unsigned int sa;                // текущий подадрес
    LOG_INFO("ioctl_man_irq_subaddress_tr");
    copy_from_user((void *) &irq_sa, (void *) arg, sizeof(INTERRUPT_SUBADDRESS)); // получаем данные из юзерспейс
    mutex_lock(&dev->mtxCh);
    switch (dev->numCh) {  // управляем прерыванием на передачу
    case CH_NUM_1:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH1, 0, rtReg, RT_TR_REG_CH1, 24, sa);
        break;
    case CH_NUM_2:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH2, 0, rtReg, RT_TR_REG_CH2, 24, sa);
        break;
    case CH_NUM_3:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH3, 0, rtReg, RT_TR_REG_CH3, 24, sa);
        break;
    case CH_NUM_4:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH4, 0, rtReg, RT_TR_REG_CH4, 24, sa);
        break;
    default:
        LOG_ERROR("bad channel");
        mutex_unlock(&dev->mtxCh);
        return ERR_BAD_CHANNEL;
    }
    mutex_unlock(&dev->mtxCh);
    return GOOD;
}

long ioctl_man_irq_subaddress_rcv(struct mil1553_cdev_data* dev, unsigned long arg)
{
    INTERRUPT_SUBADDRESS irq_sa;    // входные данные по управлению прерыванием
    unsigned int rtIntReg;          // регистр RT_INT_REG
    unsigned int rtReg;             // регистр RT_TR_REG
    unsigned int sa;                // текущий подадрес
    LOG_INFO("ioctl_man_irq_subaddress_rcv");
    copy_from_user((void *) &irq_sa, (void *) arg, sizeof(INTERRUPT_SUBADDRESS)); // получаем данные из юзерспейс
    mutex_lock(&dev->mtxCh);
    switch (dev->numCh) { // управляем прерыванием на приём
    case CH_NUM_1:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH1, 1, rtReg, RT_RCV_REG_CH1, 29, sa);
        break;
    case CH_NUM_2:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH2, 1, rtReg, RT_RCV_REG_CH2, 29, sa);
        break;
    case CH_NUM_3:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH3, 1, rtReg, RT_RCV_REG_CH3, 29, sa);
        break;
    case CH_NUM_4:
        SUBADDRESS_INTERRUPT_MANAGE(irq_sa, rtIntReg, RT_INT_REG_CH4, 1, rtReg, RT_RCV_REG_CH4, 29, sa);
        break;
    default:
        LOG_ERROR("bad channel");
        mutex_unlock(&dev->mtxCh);
        return ERR_BAD_CHANNEL;
    }
    mutex_unlock(&dev->mtxCh);
    return GOOD;
}

long ioctl_set_address(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned char address;      // адрес ОУ (1-30)
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    int shift;                  // смещение
    int countOfFigures;         // кол-во единиц
    LOG_INFO("ioctl_set_address");
    copy_from_user((void *) &address, (void *) arg, sizeof(unsigned char)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        if ((address >= 0) && (address <= 30)) {
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, CTRL_REG_PCI_CH1, CTRL_REG_PCI_CH2, CTRL_REG_PCI_CH3, CTRL_REG_PCI_CH4);
            regVal = ioread32(dev->drv->pbar + regAddr);
            regVal &= ~(0x3F00);    // обнуляем целевые биты
            regVal |= ((unsigned int)address << 8);    // записываем адрес
            countOfFigures = 0;
            for (shift = 0; shift < 6; shift++) { // считаем кол-во единиц в адресе
                if (address & (1 << shift))
                    countOfFigures++;
            }
            if (countOfFigures % 2) {   // нечётное кол-во единиц
                regVal &= ~(1 << 13);
            } else {    // чётное кол-во единиц
                regVal |=  (1 << 13);
            }
            iowrite32(regVal, dev->drv->pbar + regAddr);
        } else {
            mutex_unlock(&dev->mtxCh);
            LOG_ERROR("bad argument - address");
            return ERR_BAD_ADDRESS;
        }
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_set_timer_rtbm_conf_reg_pci(struct mil1553_cdev_data* dev, unsigned long arg)
{
    TIMER_TIMEOUT timeout;
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_set_timer_rtbm_conf_reg_pci");
    copy_from_user((void *) &timeout, (void *) arg, sizeof(TIMER_TIMEOUT)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RTBM_CONF_REG_PCI_CH1, RTBM_CONF_REG_PCI_CH2, RTBM_CONF_REG_PCI_CH3, RTBM_CONF_REG_PCI_CH4);
        // проверка tck
        switch (timeout.tck) {
        case MIL_T_TCK_OFF:
        case MIL_T_TCK_1MKS:
        case MIL_T_TCK_2MKS:
        case MIL_T_TCK_4MKS:
        case MIL_T_TCK_8MKS:
        case MIL_T_TCK_16MKS:
        case MIL_T_TCK_32MKS:
        case MIL_T_TCK_64MKS:
            LOG_INFO("tck - good");
            break;
        default:
            LOG_ERROR("bad argument - tck");
            return ERR_BAD_ARG_TCK;
        }
        // проверка trck
        switch (timeout.trck) {
        case MIL_T_TRCK_6MKS:
        case MIL_T_TRCK_8MKS:
        case MIL_T_TRCK_11MKS:
        case MIL_T_TRCK_13MKS:
        case MIL_T_TRCK_18MKS:
        case MIL_T_TRCK_61MKS:
        case MIL_T_TRCK_86MKS:
        case MIL_T_TRCK_111MKS:
            LOG_INFO("trck - good");
            break;
        default:
            LOG_ERROR("bad argument - trck");
            return ERR_BAD_ARG_TRCK;
        }
        // проверка rcvck
        switch (timeout.rcvck) {
        case MIL_T_RCVCK_17MKS:
        case MIL_T_RCVCK_60MKS:
        case MIL_T_RCVCK_85MKS:
        case MIL_T_RCVCK_110MKS:
            LOG_INFO("rcvck - good");
            break;
        default:
            LOG_ERROR("bad argument - rcvck");
            return ERR_BAD_ARG_RCVCK;
        }
        mutex_lock(&dev->mtxCh);
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal &= ~(0x70000);           // обнуляем tck
        regVal &= ~(0x700);             // обнуляем trck
        regVal &= ~(0x3);               // обнуляем rcvck
        regVal |= timeout.tck << 16;    // запись tck
        regVal |= timeout.trck << 8;    // запись trck
        regVal |= timeout.rcvck;        // запись rcvck
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_set_timer_bc_conf_reg_pci(struct mil1553_cdev_data* dev, unsigned long arg)
{
    TIMER_TIMEOUT timeout;
    unsigned int regVal;        // значение регистра
    unsigned int regAddr;       // адрес регистра
    LOG_INFO("ioctl_set_timer_rtbm_conf_reg_pci");
    copy_from_user((void *) &timeout, (void *) arg, sizeof(TIMER_TIMEOUT)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_CONF_REG_PCI_CH1, BC_CONF_REG_PCI_CH2, BC_CONF_REG_PCI_CH3, BC_CONF_REG_PCI_CH4);
        // проверка tck
        switch (timeout.tck) {
        case MIL_T_TCK_OFF:
        case MIL_T_TCK_1MKS:
        case MIL_T_TCK_2MKS:
        case MIL_T_TCK_4MKS:
        case MIL_T_TCK_8MKS:
        case MIL_T_TCK_16MKS:
        case MIL_T_TCK_32MKS:
        case MIL_T_TCK_64MKS:
            LOG_INFO("tck - good");
            break;
        default:
            LOG_ERROR("bad argument - tck");
            return ERR_BAD_ARG_TCK;
        }
        // проверка trck
        switch (timeout.trck) {
        case MIL_T_TRCK_6MKS:
        case MIL_T_TRCK_8MKS:
        case MIL_T_TRCK_11MKS:
        case MIL_T_TRCK_13MKS:
        case MIL_T_TRCK_18MKS:
        case MIL_T_TRCK_61MKS:
        case MIL_T_TRCK_86MKS:
        case MIL_T_TRCK_111MKS:
            LOG_INFO("trck - good");
            break;
        default:
            LOG_ERROR("bad argument - trck");
            return ERR_BAD_ARG_TRCK;
        }
        // проверка rcvck
        switch (timeout.rcvck) {
        case MIL_T_RCVCK_17MKS:
        case MIL_T_RCVCK_60MKS:
        case MIL_T_RCVCK_85MKS:
        case MIL_T_RCVCK_110MKS:
            LOG_INFO("rcvck - good");
            break;
        default:
            LOG_ERROR("bad argument - rcvck");
            return ERR_BAD_ARG_RCVCK;
        }
        mutex_lock(&dev->mtxCh);
        regVal = ioread32(dev->drv->pbar + regAddr);
        regVal &= ~(0x7000000);         // обнуляем tck
        regVal &= ~(0xE00000);          // обнуляем trck
        regVal &= ~(0x180000);          // обнуляем rcvck
        regVal |= timeout.tck << 24;    // запись tck
        regVal |= timeout.trck << 21;   // запись trck
        regVal |= timeout.rcvck << 19;  // запись rcvck
        iowrite32(regVal, dev->drv->pbar + regAddr);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_wr_block_buf_subadr(struct mil1553_cdev_data* dev, unsigned long arg)
{
    RT_TR_BUF_SUBADDR bufSa;        // подадрес
    unsigned int regAddr;           // адрес регистра
    int numDWord;                   // номер двойного слова аппартного буфера
    int numDataWord;                // номер слова данных программного бфуера
    LOG_INFO("ioctl_wr_block_buf_subadr");
    copy_from_user((void *) &bufSa, (void *) arg, sizeof(RT_TR_BUF_SUBADDR)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        // проверка аргументов
        if ((bufSa.num_buf == MIL_RT_BUF0) || (bufSa.num_buf == MIL_RT_BUF1)) { // проверка номера буфера
            LOG_INFO("num_buf - good");
        } else {
            LOG_ERROR("bad argument - num_buf");
            return ERR_BAD_ARG_NUMBUF;
        }
        if ((bufSa.subaddress >= 1) && (bufSa.subaddress <= 30)) { // проверка подадреса
            LOG_INFO("subaddress - good");
        } else {
            LOG_ERROR("bad argument - subaddress");
            return ERR_BAD_ARG_SA;
        }
        // запись в подадрес
        mutex_lock(&dev->mtxCh);
        if (bufSa.num_buf == MIL_RT_BUF0) { // выбор буфера
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_DATA_BUF0_CH1, RT_DATA_BUF0_CH2, RT_DATA_BUF0_CH3, RT_DATA_BUF0_CH4);
        } else {
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_DATA_BUF1_CH1, RT_DATA_BUF1_CH2, RT_DATA_BUF1_CH3, RT_DATA_BUF1_CH4);
        }
        regAddr += bufSa.subaddress * (32*2); // выбор подадреса
        for (numDWord = 0, numDataWord = 0; numDWord < 16; numDWord += 1, numDataWord += 2) { // запись слов данных в подадрес
            iowrite32(bufSa.data_words[numDataWord] | (bufSa.data_words[numDataWord + 1] << 16), dev->drv->pbar + regAddr + numDWord * 0x4);
        }
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_rd_block_buf_subadr(struct mil1553_cdev_data* dev, unsigned long arg)
{
    char isErr;                     // ошибка
    RT_TR_BUF_SUBADDR bufSa;        // подадрес
    unsigned int regAddr;           // адрес регистра
    unsigned int regVal;            // значение регистра
    int numDWord;                   // номер двойного слова аппартного буфера
    int numDataWord;                // номер слова данных программного бфуера
    LOG_INFO("ioctl_rd_block_buf_subadr");
    copy_from_user((void *) &bufSa, (void *) arg, sizeof(RT_TR_BUF_SUBADDR)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        // проверка аргументов
        if ((bufSa.num_buf == MIL_RT_BUF0) || (bufSa.num_buf == MIL_RT_BUF1)) { // проверка номера буфера
            LOG_INFO("num_buf - good");
        } else {
            LOG_ERROR("bad argument - num_buf");
            return ERR_BAD_ARG_NUMBUF;
        }
        if ((bufSa.subaddress >= 1) && (bufSa.subaddress <= 30)) { // проверка подадреса
            LOG_INFO("subaddress - good");
        } else {
            LOG_ERROR("bad argument - subaddress");
            return ERR_BAD_ARG_SA;
        }
        // запись в подадрес
        mutex_lock(&dev->mtxCh);
        if (bufSa.num_buf == MIL_RT_BUF0) { // выбор буфера
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_DATA_BUF0_CH1, RT_DATA_BUF0_CH2, RT_DATA_BUF0_CH3, RT_DATA_BUF0_CH4);
        } else {
            regAddr = CHOOSE_CHANNEL_REG(dev->numCh, RT_DATA_BUF1_CH1, RT_DATA_BUF1_CH2, RT_DATA_BUF1_CH3, RT_DATA_BUF1_CH4);
        }
        regAddr += bufSa.subaddress * (32*2); // выбор подадреса
        for (numDWord = 0, numDataWord = 0; numDWord < 16; numDWord += 1, numDataWord += 2) { // чтение слов данных из подадреса
            regVal = ioread32(dev->drv->pbar + regAddr + numDWord * 0x4);
            bufSa.data_words[numDataWord]       = regVal & 0xFFFF;
            bufSa.data_words[numDataWord + 1]   = (regVal >> 16) & 0xFFFF;
        }
        isErr = copy_to_user((void*)arg, (void*)&bufSa, sizeof(RT_TR_BUF_SUBADDR)); // возвращаем результат операции в юзерспейс
        mutex_unlock(&dev->mtxCh);
        if (isErr)
            return ERR_COPY_TO_USER;
        else
            return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_write_bc_ram(struct mil1553_cdev_data* dev, unsigned long arg)
{
    BC_RAM_BLOCK block;             // данные из юзерспейс
    unsigned int regAddr;           // адрес регистра
    unsigned int index;             // текущий индекс слова
    unsigned int* dWords;           // инструкции
    LOG_INFO("ioctl_write_bc_ram");
    copy_from_user((void *) &block, (void *) arg, sizeof(BC_RAM_BLOCK)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        // выбрать адрес куда писать в зависимости от типа
        switch (block.type) {
        case MIL_BC_RAM_TYPE_INSTRUCTION:   regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_INSTR_RAM_CH1, BC_INSTR_RAM_CH2, BC_INSTR_RAM_CH3, BC_INSTR_RAM_CH4);
            break;
        case MIL_BC_RAM_TYPE_OPERATION:     regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_OPERATION_RAM_CH1, BC_OPERATION_RAM_CH2, BC_OPERATION_RAM_CH3, BC_OPERATION_RAM_CH4);
            break;
        case MIL_BC_RAM_TYPE_DATA:          regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_DATA_RAM_CH1, BC_DATA_RAM_CH2, BC_DATA_RAM_CH3, BC_DATA_RAM_CH4);
            break;
        default:
            LOG_ERROR("bad argument - type = %d", block.type);
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_ARG_TYPE;
        }
        // выделяем память для инструкций
        dWords = kmalloc(block.length * 4, GFP_KERNEL);
        if (!dWords)
        {
            LOG_ERROR("bad alloc mem for dwords");
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_ALLOC_MEM;
        }
        // копируем инструкции из юзерспейс
        copy_from_user((void *) dWords, (void *) (block.dwords), block.length * 4);
        // добавляем смещенние
        regAddr += block.shift * 4;
        // пословная запись слов в bc ram
        for (index = 0; index < block.length; index++) {
            iowrite32(dWords[index], dev->drv->pbar + regAddr);
            regAddr += 4;
            //LOG_INFO("wrt word = 0x%x at 0x%x", dWords[index], regAddr + (block.shift + index) * 4);
        }
        kfree(dWords);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_read_bc_ram(struct mil1553_cdev_data* dev, unsigned long arg)
{
    BC_RAM_BLOCK block;             // данные из юзерспейс
    unsigned int regAddr;           // адрес регистра
    unsigned int index;             // текущий индекс слова
    unsigned int* dWords;           // инструкции
    LOG_INFO("ioctl_read_bc_ram");
    copy_from_user((void *) &block, (void *) arg, sizeof(BC_RAM_BLOCK)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        // выбрать адрес откуда читать в зависимости от типа
        switch (block.type) {
        case MIL_BC_RAM_TYPE_INSTRUCTION:   regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_INSTR_RAM_CH1, BC_INSTR_RAM_CH2, BC_INSTR_RAM_CH3, BC_INSTR_RAM_CH4);
            break;
        case MIL_BC_RAM_TYPE_OPERATION:     regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_OPERATION_RAM_CH1, BC_OPERATION_RAM_CH2, BC_OPERATION_RAM_CH3, BC_OPERATION_RAM_CH4);
            break;
        case MIL_BC_RAM_TYPE_DATA:          regAddr = CHOOSE_CHANNEL_REG(dev->numCh, BC_DATA_RAM_CH1, BC_DATA_RAM_CH2, BC_DATA_RAM_CH3, BC_DATA_RAM_CH4);
            break;
        default:
            LOG_ERROR("bad argument - type = %d", block.type);
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_ARG_TYPE;
        }
        // добавляем смещенние
        regAddr += block.shift * 4;
        // выделяем память для инструкций
        dWords = kmalloc(block.length * 4, GFP_KERNEL);
        if (!dWords)
        {
            LOG_ERROR("bad alloc mem for dwords");
            mutex_unlock(&dev->mtxCh);
            return ERR_BAD_ALLOC_MEM;
        }
        // пословное чтение слов из bc ram
        for (index = 0; index < block.length; index++) {
            dWords[index] = ioread32(dev->drv->pbar + regAddr);
            regAddr += 4;
            //LOG_INFO("rd word = 0x%x at 0x%x", dWords[index], regAddr + (block.shift + index) * 4);
        }
        // копируем в юзерспейс блок
        copy_to_user((void*)block.dwords, (void*)dWords, block.length * 4);
        kfree(dWords);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_get_count_interrupt_blocks(struct mil1553_cdev_data* dev, unsigned long arg)
{
    char isErr;         // ошибка
    unsigned int count; // кол-во блоков, доступных для чтения
    LOG_INFO("ioctl_get_count_interrupt_blocks");
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        LOG_INFO("indRead = %d, indWrt = %d", dev->interrupt_buffer->index_read, dev->interrupt_buffer->index_write);
        if ( dev->interrupt_buffer->index_read <= dev->interrupt_buffer->index_write ) {
            count = dev->interrupt_buffer->index_write - dev->interrupt_buffer->index_read;
            LOG_INFO("if - count = %d", count);
        } else {
            count = (dev->interrupt_buffer->count_blocks - dev->interrupt_buffer->index_read) + dev->interrupt_buffer->index_write;
            LOG_INFO("else - count = %d", count);
        }
        isErr = copy_to_user((void*)arg, (void*)&count, sizeof(unsigned int));
        mutex_unlock(&dev->mtxCh);
        if (isErr)
            return ERR_COPY_TO_USER;
        else
            return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_read_interrupt_blocks(struct mil1553_cdev_data* dev, unsigned long arg)
{
    unsigned int nblock;        // кол-во блоков, доступных для чтения
    unsigned int nblockend;     // кол-во блоков, до границы буфера
    int err;
    unsigned int count_blocks;
    struct block_info*  blocks = 0;
    //INTERRUPT_BLOCK_BUFFER block;       // блок данных
    err = copy_from_user((void *) &count_blocks, (void *) arg, sizeof(unsigned int)); // получаем данные из юзерспейс
    err = copy_from_user((void*) &blocks, (void*)(arg + 4), sizeof(unsigned long)); // получаем данные из юзерспейс
    LOG_INFO("ioctl_read_interrupt_blocks - cnt blocks = %d", count_blocks);
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        mutex_lock(&dev->mtxCh);
        if (dev->interrupt_buffer->index_write >= dev->interrupt_buffer->index_read) {
            nblock = dev->interrupt_buffer->index_write - dev->interrupt_buffer->index_read;
        } else {
            nblock = dev->interrupt_buffer->count_blocks - dev->interrupt_buffer->index_read + dev->interrupt_buffer->index_write;
        }
        if (count_blocks > nblock)
            count_blocks = nblock;
        nblockend = dev->interrupt_buffer->count_blocks - dev->interrupt_buffer->index_read; // вычисляем кол-во блоков, до границы буфера
        if (count_blocks <= nblockend) { // кол-во запрошенных блоков не превышает границу буфера
            LOG_INFO("block.count_blocks <= nblockend");
            err = copy_to_user((void*)blocks, (void*)&(dev->interrupt_buffer->blocks[dev->interrupt_buffer->index_read]), count_blocks * sizeof(struct block_info)); // копируем в юзерспейс блоки
            dev->interrupt_buffer->index_read += count_blocks; // сдвигаем указатель чтения
            if (dev->interrupt_buffer->index_read >= dev->interrupt_buffer->count_blocks)
                dev->interrupt_buffer->index_read = 0; // при необходимости корректируем
        } else {
            LOG_INFO("block.count_blocks > nblockend 1, nblockend = %d, nblock = %d, cntblks = %d", nblockend, nblock, count_blocks);
            err = copy_to_user((void*)blocks, (void*)&(dev->interrupt_buffer->blocks[dev->interrupt_buffer->index_read]), nblockend * sizeof(struct block_info));  // копируем первый кусок блоков
            dev->interrupt_buffer->index_read = 0;
            LOG_INFO("block.count_blocks > nblockend 2");
            err = copy_to_user((void*)(blocks + nblockend * sizeof(struct block_info)), (void*)&(dev->interrupt_buffer->blocks[dev->interrupt_buffer->index_read]), (count_blocks - nblockend) * sizeof(struct block_info)); // копируем второй кусок блоков
            dev->interrupt_buffer->index_read = (count_blocks - nblockend);
        }
        err = copy_to_user((void*)arg, (void*)&count_blocks, sizeof(unsigned int));  // копируем в юзерспей фактическое кол-во блоков полученных
        LOG_INFO("cnt blocks - %d", err);
        mutex_unlock(&dev->mtxCh);
        return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

long ioctl_get_rcv_subadr_data(struct mil1553_cdev_data* dev, unsigned long arg)
{
    char isErr;                         // ошибка
    SUBA_DATA_BLOCK block;              // блок поддадреса
    LOG_INFO("ioctl_get_rcv_subadr_data");
    copy_from_user((void *) &block, (void *) arg, sizeof(SUBA_DATA_BLOCK)); // получаем данные из юзерспейс
    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
        if ((block.sa >= 1) && (block.sa <= 30)) {
            LOG_INFO("sa - good");
        } else {
            LOG_ERROR("bad argument - block.sa");
            return -EFAULT;
        }
        mutex_lock(&dev->mtxCh);
        memcpy( (void*)block.words, (void*)dev->rcv_sa_buffer->data[block.sa - 1], 32 * 2 );
        isErr = copy_to_user((void*)arg, (void*) &block, sizeof(SUBA_DATA_BLOCK));
        mutex_unlock(&dev->mtxCh);
        if (isErr)
            return ERR_COPY_TO_USER;
        else
            return GOOD;
    } else {
        LOG_ERROR("bad channel");
        return ERR_BAD_CHANNEL;
    }
}

//long ioctl_set_signal_irq(struct mil1553_cdev_data* dev, unsigned long arg)
//{
//    SIGNAL_INFO info;       // инфо о сигнале
//    LOG_INFO("ioctl_set_signal_irq");
//    copy_from_user((void *) &info, (void *) arg, sizeof(SIGNAL_INFO)); // получаем данные из юзерспейс
//    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
//        if ((info.sig_number == SIGUSR1) || (info.sig_number == SIGUSR2)) {
//            LOG_INFO("sig_number - good");
//        } else {
//            LOG_ERROR("bad argument - info.sig_number");
//            return -EFAULT;
//        }
//        mutex_lock(&dev->mtxCh);
//        dev->sig_info.sig_num = info.sig_number;
//        dev->sig_info.target_pid = info.target_pid;
//        dev->sig_info.sig_task = pid_task( find_vpid(dev->sig_info.target_pid), PIDTYPE_PID );
//        dev->sig_info.isCanSend = 1;
//        mutex_unlock(&dev->mtxCh);
//        return 0;
//    } else {
//        LOG_ERROR("bad channel");
//        return -EFAULT;
//    }
//}

//long ioctl_clr_signal_irq(struct mil1553_cdev_data* dev, unsigned long arg)
//{
//    LOG_INFO("ioctl_clr_signal_irq");
//    if ((dev->numCh == CH_NUM_1) || (dev->numCh == CH_NUM_2) || (dev->numCh == CH_NUM_3) || (dev->numCh == CH_NUM_4)) {
//        mutex_lock(&dev->mtxCh);
//        dev->sig_info.sig_num = 0;
//        dev->sig_info.target_pid = 0;
//        dev->sig_info.sig_task = 0;
//        dev->sig_info.isCanSend = 0;
//        mutex_unlock(&dev->mtxCh);
//        return 0;
//    } else {
//        LOG_ERROR("bad channel");
//        return -EFAULT;
//    }
//}

long ioctl_get_number_ch(struct mil1553_cdev_data* dev, unsigned long arg)
{
    char isErr;         // ошибка
    unsigned int numCh; // номер канала
    LOG_INFO("ioctl_get_number_ch");
    mutex_lock(&dev->mtxCh);
    numCh = dev->numCh;
    isErr = copy_to_user((void*)arg, (void*) &numCh, sizeof(unsigned int));
    mutex_unlock(&dev->mtxCh);
    if (isErr)
        return ERR_COPY_TO_USER;
    else
        return GOOD;
}

