/// \file mil1553ud_pci.h
/// \brief инфраструктура для работы с pci устройством
#ifndef MIL1553UD_PCI_H
#define MIL1553UD_PCI_H

#include "mil1553ud_info.h"

#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/kthread.h>
#include <linux/completion.h>
#include <linux/sched.h>
#include <linux/semaphore.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/uaccess.h>
#include <linux/mutex.h>
#include <asm/msr.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/version.h>

/// \brief кол-во каналов 1
#define COUNT_CHANNELS_1        1
/// \brief кол-во каналов 2
#define COUNT_CHANNELS_2        2
/// \brief кол-во каналов 1
#define COUNT_CHANNELS_4        4

/// \brief идентификатор pci-устройтва ud1
#define VAL_UD1                 0x03
/// \brief идентификатор pci-устройтва ud2
#define VAL_UD2                 0x23
/// \brief идентификатор pci-устройтва ud4
#define VAL_UD4                 0x43

///
/// \brief The DMA_INFO struct информация для взаимодействия с ДМА
///
struct DMA_INFO {
    /// \brief dma_size размер примаппленной области ДМА
    size_t                  dma_size;
    /// \brief dma_paddr указатель на память ДМА
    dma_addr_t              dma_paddr;
    /// \brief dma_vaddr виртуальный указатель намять ДМА
    uint8_t*                dma_vaddr;
    /// \brief ch_dma_vaddr указатели на область ДМА - поканально
    uint8_t*                ch_dma_vaddr[MAX_CHANNELS];
    /// \brief ch_dma_read текущий указатель на чтение данных поканально
    uint32_t*               ch_dma_read[MAX_CHANNELS];
    /// \brief ch_dma_write текущий указатель на запись данных поканально
    uint32_t*               ch_dma_write[MAX_CHANNELS];
    /// \brief ch_dma_read_irq текущий указатель на чтение данных поканально (в прерывании)
    uint32_t*               ch_dma_read_irq[MAX_CHANNELS];
    /// \brief ch_dma_write_irq текущий указатель на запись данных поканально (в прерывании)
    uint32_t*               ch_dma_write_irq[MAX_CHANNELS];
};

///
/// \brief The MIL1553UD_DEV struct базовая структура для работы с pci устройтсвом
///
struct MIL1553UD_DEV {
    /// \brief pbar указатель на примаппленную память устройства
    u8 __iomem*             pbar;
    /// \brief barlen размер примаппленной памяти
    uint64_t                barlen;
    /// \brief bar адрес реcурса ввода/вывода
    int                     bar;
    /// \brief pdevice указатель на базовую pci_dev
    struct pci_dev*         pdevice;
    /// \brief unit номер pci устройства в массиве pdevs
    u_int                   unit;
    /// \brief devname имя pci устройства
    char                    devname[30];
    /// \brief vendorid идентификатор вендора
    uint16_t                vendorid;
    /// \brief deviceid идентификатор устройства
    uint16_t                deviceid;
    /// \brief revisionid идентификатор ревизии
    uint16_t                revisionid;
    /// \brief dma_info информация о ДМА
    struct DMA_INFO         dma_info;
    /// \brief mtx  мьютекс, для регулировки доступа из символьных устройств каналов к регистрам управления
    struct mutex            mtx;
    struct mutex            mtxIrq;
    /// \brief current_count_of_channels фактическое количество каналов в плате (1-4)
    int                     current_count_of_channels;
    /// \brief mil1553cdev_class класс символьного устройства канала
    struct class*           mil1553cdev_class;
    /// \brief mil1553cdev_name имя класса символьного устройства канала
    char                    mil1553cdev_name[30];
    /// \brief index_cdev индексы символьных устройств каналов pci-платы
    int                     index_cdev[MAX_CHANNELS];
};

#endif // MIL1553UD_PCI_H
